from Models.BWM.GaussianBWM import *
from Models.BWM.IntervalBWM import *
from Models.BWM.StandardBWM import *
from Models.BWM.TriangularBWM import *
from Models.BWM.BetaBWM import *
import numpy as np

a_b =  np.array([
    [ 3, 4, 6, 1, 5, 2, 9, 7],
    [ 1, 2, 8, 4, 5, 3, 9, 6],
    [ 2, 2, 3, 1, 5, 5, 9, 8],
    [ 2, 1, 8, 2, 9, 3, 8, 8],
    [ 2, 4, 9, 1, 4, 3, 5, 5],
    [ 1, 2, 9, 1, 3, 5, 5, 4],
    ])

a_w =  np.array([
    [ 7, 6, 4, 9, 5, 8, 1, 3],
    [ 9, 8, 2, 5, 4, 5, 1, 3],
    [ 8, 8, 5, 9, 5, 5, 1, 2],
    [ 8, 9, 2, 8, 1, 8, 2, 2],
    [ 8, 6, 1, 9, 6, 7, 4, 4],
    [ 9, 8, 1, 9, 7, 5, 5, 6],
    ])


## Standard BWM 
#bwm = StandardBWM( AB=a_b, AW=a_w)
#bwm.sampling()

## Beta BWM 
bwm = BetaBWM( AB_md=a_b, AB_concentration=10*np.ones(a_w.shape), AW_md=a_w, AW_concentration=10*np.ones(a_w.shape))
bwm.sampling()

## Triangular BWM 
bwm_tri = TriangularBWM( AB_md=a_b, AW_md=a_w)
#bwm_tri.sampling()

## Interval BWM 
bwm_intv = IntervalBWM(AB_l = a_b, AB_h = a_b, AW_l = a_w, AW_h = a_w)
#bwm_intv.sampling()

## Gaussian BWM 
eps = 1
bwm_gauss = GaussianBWM(AB_md = a_b, AB_sigma=eps*np.ones(a_w.shape), AW_md = a_w, AW_sigma = eps*np.ones(a_w.shape), num_samples=1000)
bwm_gauss.sampling()

print('Ok')



from ..MCDMProblem import MCDMProblem
import numpy as np

class MultinomialBWM(MCDMProblem):
    _basic_model = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            int AB[DmNo, CNo];
            int AW[DmNo, CNo];
            vector<lower=0,upper=1>[CNo] e;
        }

        parameters {
            simplex[CNo] W[DmNo];
            simplex[CNo] wStar;
            real<lower=0> kappaStar;
        }

        model {
            kappaStar ~ gamma(.001,.01);
            wStar ~ dirichlet(0.01*e);

            for (i in 1:DmNo){
                W[i] ~ dirichlet(kappaStar*wStar);
                AW[i,:] ~ multinomial(W[i]);

                vector[CNo] wInv;

                wInv = e ./ W[i];
                wInv ./= sum(wInv);
                AB[i,:] ~ multinomial(wInv);
            }
        }
    """
   
    _basic_model_clustering = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            int AB[DmNo, CNo];
            int AW[DmNo, CNo];
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> DmC;
        }

        parameters {
            simplex[CNo] W[DmNo];
            simplex[CNo] wc[DmC];
            real<lower=0> ksi[DmC];
            simplex[DmC] theta[DmNo];
        }

        model {
            ksi ~ gamma(.001, .001);
            for(i in 1:DmC)
                wc[i] ~ dirichlet(0.001*e);

            
            for (i in 1:DmNo) {
                real contribution[DmC];
                for(j in 1:DmC)
                    contribution[j] = log(theta[i,j]) + dirichlet_lpdf( W[i] | ksi[j]*wc[j]);
                target += log_sum_exp(contribution);

                AW[i,:] ~ multinomial(W[i]);

                vector[CNo] wInv;
                wInv = e ./ W[i];
                wInv ./= sum(wInv);
                AB[i,:] ~ multinomial(wInv);
            }
        }
    """

    _basicModelSorting = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            int AB[DmNo, CNo];
            int AW[DmNo, CNo];
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> AltNo;
            int<lower=2> AltC;
            matrix[AltNo, CNo] Alt;
        }

        parameters {
            simplex[CNo] W[DmNo];
            simplex[CNo] wStar;
            real<lower=0> kappaStar;

            vector[AltC] altMu;
        }

        transformed parameters {
            vector[AltNo] v = Alt * wStar;
            v ./= (1-v);
            v = log(v);
            array[AltNo, AltC] real<upper=0> soft_z; // log unnormalized clusters
            for (n in 1:AltNo)
                for (k in 1:AltC)
                    soft_z[n, k] = -log(AltC) - 0.5 * pow(altMu[k] - v[n],2);

        }

        model {
            wStar ~ dirichlet(0.01*e);
            kappaStar ~ gamma(.01,.01);

            for (i in 1:DmNo){
                W[i] ~ dirichlet(kappaStar*wStar);

                AW[i,:] ~ multinomial(W[i]);

                vector[CNo] wInv;
                wInv = e ./ W[i];
                wInv ./= sum(wInv);
                AB[i,:] ~ multinomial(wInv);
            }

            for (n in 1:AltNo)
                target += log_sum_exp(soft_z[n]);
        }
    """

    _correlatedModel = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            int AB[DmNo, CNo];
            int AW[DmNo, CNo];
            vector<lower=0,upper=1>[CNo] e;

            vector[CNo] mu; // mean of unnormalized weight
            cov_matrix[CNo] Sigma;
        }

        parameters {
            vector[CNo] W_eta[DmNo];
            vector[CNo] wStar_eta;
            real<lower=0> kappaStar;
        }

        transformed parameters {
            simplex[CNo] W[DmNo];
            simplex[CNo] wStar;

            wStar = softmax(wStar_eta);

            for(i in 1:DmNo)
                W[i] = softmax(W_eta[i]);
        }

        model {
            wStar_eta ~ multi_normal(mu, Sigma);

            for (i in 1:DmNo){
                W_eta[i] ~ multi_normal(wStar_eta, 0.1*Sigma);

                AW[i,:] ~ multinomial(W[i]);

                vector[CNo] wInv;
                wInv = e ./ W[i];
                wInv ./= sum(wInv);
                AB[i,:] ~ multinomial(wInv);
            }
        }
    """

    _correlatedModelClustering = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            int AB[DmNo, CNo];
            int AW[DmNo, CNo];
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> DmC;

            vector[CNo] mu; // mean of unnormalized weight
            cov_matrix[CNo] Sigma;
        }

        parameters {
            vector[CNo] W_eta[DmNo];
            vector[CNo] wc_eta[DmC];
            real<lower=0> ksi[DmC];
            simplex[DmC] theta[DmNo];
        }

        transformed parameters {
            simplex[CNo] W[DmNo];
            simplex[CNo] wc[DmC];

            for(i in 1:DmNo)
                W[i] = softmax(W_eta[i]);

            for(i in 1:DmC)
                wc[i] = softmax(wc_eta[i]);
        }

        model {
            real contribution[DmC];
            ksi ~ gamma(.001, .001);

            for (d in 1:DmC)
                wc_eta[d] ~ multi_normal(mu, Sigma);//wc_eta[d] ~ dirichlet(.001*e);//

            for (i in 1:DmNo){
                for(j in 1:DmC)
                    contribution[j] = log(theta[i,j]) + multi_normal_lpdf( W_eta[i] | wc_eta[j], Sigma*.01);
                    //contribution[j] = log(theta[i,j]) + dirichlet_lpdf( W[i] | ksi[j]*wc[j]);

                target += log_sum_exp(contribution);


                AW[i,:] ~ multinomial(W[i]);
                vector[CNo] wInv;
                wInv = e ./ W[i];
                wInv ./= sum(wInv);
                AB[i,:] ~ multinomial(wInv);
            }
        }
    """

    _correlatedModelSorting = """
        data {
            int<lower=2> CNo;
            int<lower=1> DmNo;
            int AB[DmNo, CNo];
            int AW[DmNo, CNo];
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> AltNo;
            matrix[AltNo, CNo] Alt;
            int<lower=2> AltC;
            vector<lower=0,upper=1>[AltC] eAlt;

            vector[CNo] mu; // mean of unnormalized weight
            cov_matrix[CNo] Sigma;
        }

        parameters {
            vector[CNo] W_eta[DmNo];
            vector[CNo] wStar_eta;
            real<lower=0> kappaStar;

            vector[AltC] altMu;
        }

        transformed parameters {
            simplex[CNo] W[DmNo];
            simplex[CNo] wStar;

            wStar = softmax(wStar_eta);

            for(i in 1:DmNo)
                W[i] = softmax(W_eta[i]);


            vector[AltNo] v = Alt * wStar;
            v ./= (1-v);
            v = log(v);
            array[AltNo, AltC] real<upper=0> soft_z; // log unnormalized clusters

            for (n in 1:AltNo)
                for (k in 1:AltC)
                    soft_z[n, k] = -log(AltC) - 0.5 * pow(altMu[k] - v[n],2);

        }

        model {
            kappaStar ~ gamma(.0001,.001);
            wStar_eta ~ multi_normal(mu, Sigma);

            for (i in 1:DmNo){
                W_eta[i] ~ multi_normal(wStar_eta, 0.01*Sigma);

                AW[i,:] ~ multinomial(W[i]);

                vector[CNo] wInv;
                wInv = e ./ W[i];
                wInv ./= sum(wInv);
                AB[i,:] ~ multinomial(wInv);
            }

            for (n in 1:AltNo)
                target += log_sum_exp(soft_z[n]);

        }
    """

    def __init__(self, AB, AW, alternatives = None, dm_cluster_number=-1, alt_sort_number=-1, num_chain=3, num_samples=1000, opt={}):
        self.AB = np.array(AB)
        self.AW = np.array(AW)

        super().__init__(alternatives, dm_cluster_number, alt_sort_number, num_chain, num_samples, opt)


    @property
    def inputData(self):
        data = self._getCommonData()
        data['AW'] = self.AW
        data['AB'] = self.AB
        data['e'] =  np.ones(self.CNo)

        return data

    @property
    def OriginalModel(self):
        return self.__originalModel

    @property
    def DmNo(self):
        return self.AB.shape[0]

    @property
    def CNo(self):
        return self.AB.shape[1]

    def _checkInputData(self):
        assert self.AB.shape == self.AW.shape, "AB and AW must be of the same size!"
        assert self.AB.shape[0] >= 1, "No input"
        assert self.AW.shape[1] >= 2, "The number of criteria must be more than 2!"

        return True


if __name__ == "__main__":
    a_b =  np.array([
       [ 3, 4, 6, 1, 5, 2, 9, 7],
       [ 1, 2, 8, 4, 5, 3, 9, 6],
       [ 2, 2, 3, 1, 5, 5, 9, 8],
       [ 2, 1, 8, 2, 9, 3, 8, 8],
       [ 2, 4, 9, 1, 4, 3, 5, 5],
       [ 1, 2, 9, 1, 3, 5, 5, 4]])

    a_w =  np.array([
        [ 7, 6, 4, 9, 5, 8, 1, 3],
        [ 9, 8, 2, 5, 4, 5, 1, 3],
        [ 8, 8, 5, 9, 5, 5, 1, 2],
        [ 8, 9, 2, 8, 1, 8, 2, 2],
        [ 8, 6, 1, 9, 6, 7, 4, 4],
        [ 9, 8, 1, 9, 7, 5, 5, 6]])

    dmNo, cNo = a_w.shape
    altNo = 50
    x = np.random.rand(altNo // 2, cNo)
    altMat = np.concatenate([x*1,x])

    opt = {'CriteriaDependence': False, 'Sigma': np.eye(cNo) }

    bwm = MultinomialBWM( AB=a_b, AW=a_w, opt=opt, alternatives=altMat)#, dm_cluster_number=2, )
    bwm.sampling()
    print('Ok')




import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
from scipy import stats

class WeightVisualizer:
    def __init__(self, weight_samples, criteria_names):
        """
        weight_samples: numpy array of shape (n_samples, n_criteria)
        criteria_names: list of criteria names (length n_criteria)
        """
        self.weight_samples = weight_samples
        self.criteria_names = criteria_names

    def plot_distributions(self, **kwargs):
        """
        Plots the distributions of the weight samples for each criterion.

        kwargs:
            ncols: number of columns in the subplot grid (default: 3)
            xlim: tuple for x-axis limits (default: (0, 1))
            xticks: list or array for x-axis ticks (default: None)
            font_size: font size for titles and labels (default: 12)
            font_bold: whether to use bold font for titles and labels (default: True)
        """
        import matplotlib.pyplot as plt

        ncols = kwargs.get('ncols', 3)
        xlim = kwargs.get('xlim', (0, 1))
        xticks = kwargs.get('xticks', None)
        font_size = kwargs.get('font_size', 12)
        font_bold = kwargs.get('font_bold', True)

        fontweight = 'bold' if font_bold else 'normal'

        c_no = len(self.criteria_names)
        cols = ncols
        rows = int(np.ceil(c_no / cols))

        fig, axs = plt.subplots(rows, cols, figsize=(
            4 * cols, 3 * rows), sharey=True, sharex=True)
        axs = axs.flatten() if c_no > 1 else [axs]

        for idx in range(rows * cols):
            if idx < c_no:
                axs[idx].hist(self.weight_samples[:, idx], bins=50)
                mean_val = np.mean(self.weight_samples[:, idx])
                axs[idx].axvline(mean_val, color='red',
                                 linestyle='--', linewidth=2)
                axs[idx].set_title(
                    f"{self.criteria_names[idx]} (mean={mean_val:.3f})",
                    fontsize=font_size,
                    fontweight=fontweight
                )
                if xticks is not None:
                    axs[idx].set_xticks(xticks)
                axs[idx].set_xlim(xlim)
                axs[idx].tick_params(axis='both', labelsize=font_size)
                for label in (axs[idx].get_xticklabels() + axs[idx].get_yticklabels()):
                    label.set_fontweight(fontweight)
            else:
                fig.delaxes(axs[idx])

        plt.tight_layout()
        plt.show()

    def ridge_plot(self, **kwargs):
        """
        Plots the posterior weight distributions for each criterion using seaborn FacetGrid.

        kwargs:
            font_size: font size for titles and labels (default: 12)
            font_bold: whether to use bold font for titles and labels (default: True)
            palette: color palette for seaborn (default: "viridis")
        """
        font_size = kwargs.get('font_size', 14)
        font_bold = kwargs.get('font_bold', True)
        palette = kwargs.get('palette', "viridis")

        fontweight = 'bold' if font_bold else 'normal'

        # Convert to long-form DataFrame
        df_long = pd.DataFrame(self.weight_samples, columns=self.criteria_names).melt(
            var_name="Criterion", value_name="Weight"
        )

        # Calculate means for each criterion
        means = {name: self.weight_samples[:, idx].mean()
                 for idx, name in enumerate(self.criteria_names)}

        # Set plot style
        sns.set(style="white", rc={"axes.facecolor": (0, 0, 0, 0)})

        # Initialize the FacetGrid
        g = sns.FacetGrid(df_long, row="Criterion", hue="Criterion",
                          aspect=10, height=1.2, palette=palette)
        # Draw densities
        g.map(sns.kdeplot, "Weight",
              bw_adjust=1, clip_on=False, fill=True, alpha=0.8, linewidth=1.5)

        # Add a horizontal line at zero
        g.map(plt.axhline, y=0, lw=1, clip_on=False)

        # Add criteria names and means as text on each facet
        for ax, criterion in zip(g.axes.flat, self.criteria_names):
            mean_val = means[criterion]
            ax.text(
                0.98, 0.7,
                f"{criterion} (mean={mean_val:.3f})",
                transform=ax.transAxes,
                ha='right', va='center', fontsize=font_size, fontweight=fontweight, color='black'
            )
            # Plot red dotted line at mean
            ax.axvline(mean_val, color='red', linestyle='--',
                       linewidth=2, alpha=0.8)

        # Set titles and layout
        g.figure.subplots_adjust(hspace=-0.4)
        g.set_titles("")
        g.set(yticks=[], ylabel="")
        g.despine(bottom=True, left=True)

        plt.xlabel("Posterior Weight", fontsize=font_size, fontweight=fontweight)
        plt.tight_layout()
        plt.show()

    from scipy import stats

    def vertical_jakplot(self, **kwargs):
        """
        Plot vertical jakplot-style distributions for each criterion's weights.

        Parameters:
            samples (np.ndarray): shape (n_samples, n_criteria)
            criteria_names (list): list of criterion names
            font_size (int, optional): font size for labels and ticks
            font_bold (bool, optional): whether to use bold font
            ylim (tuple, optional): y-axis limits (min, max)
        """
        font_size = kwargs.get('font_size', 12)
        font_bold = kwargs.get('font_bold', False)
        ylim = kwargs.get('ylim', None)

        label_weight = 'bold' if font_bold else 'normal'

        n_criteria = self.weight_samples.shape[1]
        plt.figure(figsize=(7, 4))
        for idx in range(n_criteria):
            data = self.weight_samples[:, idx]
            # Estimate density using KDE
            kde = stats.gaussian_kde(data)
            y_grid = np.linspace(data.min(), data.max(), 300)
            pdf = kde(y_grid)
            # Normalize pdf for better visualization
            pdf = pdf / pdf.max() * 0.4  # scale width
            plt.plot(idx + pdf, y_grid, color='black', linewidth=3)
            plt.plot([idx], [data.mean()], 'o', color='blue')
        plt.xticks(range(n_criteria), self.criteria_names, fontsize=font_size, fontweight=label_weight)
        plt.xlabel('Criterion', fontsize=font_size, fontweight=label_weight)
        plt.ylabel('Weight', fontsize=font_size, fontweight=label_weight)
        if ylim is not None:
            plt.ylim(ylim)
        plt.tight_layout()
        plt.show()
    


from Models.BWM.StandardBWM import *
import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from scipy.special import softmax

df = pd.read_csv('toursim.csv')
dmNo = df.shape[0]
cNo = int(df.shape[1]/2)

a_b = df.values[:,:cNo]
a_w = df.values[:, cNo:]

altNo = 50
x = np.random.rand(altNo // 2, cNo)
altMat = np.concatenate([x*1,x])

cluster_no = 3
## Criteria independence 
opt = {'CriteriaDependence': False }
bwm = StandardBWM( AB=a_b, AW=a_w, opt=opt, dm_cluster_number=cluster_no, num_samples=1000)
bwm.sampling()

weights = bwm.DmWeight
geo_mean = np.exp(np.mean(np.log(weights), axis=1))
clr_weights = np.log(weights / geo_mean[:, None])
kmeans = KMeans(n_clusters=cluster_no, random_state=0).fit(clr_weights)
centers = kmeans.cluster_centers_
centers_compositional = softmax(centers, axis=1)

print('Ok')


from .BWM.StandardBWM import StandardBWM

# /Users/majidmohammadi/surfdrive/Research/MCDM/BayesianMCDM/Models/__init__.py

# Redirect BayesMCDM.BWM.StandardBWM to Models.BWM.StandardBWM.StandardBWM


# Optionally, provide a namespace for BWM if needed
class BWM:
    StandardBWM = StandardBWM


from pyvis.network import Network
import matplotlib.pyplot as plt
import numpy as np
from IPython.display import HTML
import networkx as nx
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize
import networkx as nx
from matplotlib.lines import Line2D
import numpy as np
import matplotlib.cm as cmx

from graphviz import Digraph
from IPython.display import display

class CredalRanking:
    def __init__(self, weight_samples, **kwargs):
        """
        Initializes the CredalRanking object.

        Parameters
        ----------
        weight_samples : np.ndarray
            A 2D numpy array of shape (num_samples, num_criteria) containing sampled weights for each criterion.
        criteria_name : list of str, optional
            List of criterion names. If not provided, defaults to ['C1', 'C2', ..., 'Cn'] where n = number of criteria.

        Raises
        ------
        AssertionError
            If the length of criteria_name does not match the number of criteria in weight_samples.
        """

        if 'criteria_name' in kwargs:
            self.criteria_names = kwargs['criteria_names'].copy()
        else:
            self.criteria_names = [f'C{i+1}' for i in range(weight_samples.shape[1])]

        assert len(self.criteria_names) == weight_samples.shape[1], "Invalid number of criteria"

        self.weight_samples = weight_samples
        self.avg_weights = np.mean(self.weight_samples, axis=0)
        self.index = np.argsort(-self.avg_weights)
        assert len(self.criteria_names) == self.weight_samples.shape[1], "Invalid number of criteria"
        self.sample_no, self.c_no = self.weight_samples.shape
        # for i in range(self.c_no):
        #     self.criteria_names[i] = self.criteria_names[i] + ' - ' + str(round(self.avg_weights[i],3))
        self.probs = np.empty((self.c_no, self.c_no))
        for i in range(self.c_no):
            for j in range(i, self.c_no):
                self.probs[i,j] = round((self.weight_samples[:,i] >= self.weight_samples[:,j]).sum() / self.sample_no,2)
                self.probs[j,i] = 1 - self.probs[i,j]

        # Edge styles as a class property
        self.edge_styles = [
            {'range': (0.5, 0.6), 'color': 'lightblue', 'style': 'dotted', 'label': '0.5–0.6', 'width': 1},
            {'range': (0.6, 0.7), 'color': 'deepskyblue', 'style': 'dashdot', 'label': '0.6–0.7', 'width': 2},
            {'range': (0.7, 0.8), 'color': 'dodgerblue', 'style': 'dashed', 'label': '0.7–0.8', 'width': 2.5},
            {'range': (0.8, 0.9), 'color': 'blue', 'style': 'solid', 'label': '0.8–0.9', 'width': 2.5},
            {'range': (0.9, 1.01), 'color': 'navy', 'style': 'solid', 'label': '0.9–1.0', 'width': 3}
        ]
    
    def plot(self, **kwargs):
        """
        Visualizes the credal ranking of criteria using a directed graph.
        ----------
        show_weight : bool, optional
            If True, displays the probability value on each edge. Default is False.
        **kwargs
            Additional keyword arguments for future extensions.
        Notes
        -----
        - Nodes are positioned in a circle for clarity, with labels placed outside the circle.
        - If show_weight is True, edge weights are displayed as labels.
        - Otherwise, Edge styles (color, width, linestyle) are determined by the probability intervals in `self.edge_styles`.
        - A legend is included to explain edge styles unless `show_weight` is True.
        """

        show_weight = kwargs.get('show_weight', False)
        avg_weights = self.avg_weights
        # criteria_names = [name.split(' - ')[0] for name in self.criteria_names]
        node_labels = {name: f"{name}\n({avg_weights[i]:.2f})" for i, name in enumerate(self.criteria_names)}

        # Create directed graph with criteria_names as nodes
        G = nx.DiGraph()
        G.add_nodes_from(self.criteria_names)

        # Add edges based on self.probs
        for i, src in enumerate(self.criteria_names):
            for j, tgt in enumerate(self.criteria_names):
                if i != j:
                    prob = self.probs[i, j]
                    for style in self.edge_styles:
                        lo, hi = style['range']
                        if lo <= prob < hi:
                            G.add_edge(src, tgt, weight=prob, color=style['color'], style=style['style'])
                            break

        # Sort nodes by descending average weight
        sorted_indices = np.argsort(-avg_weights)
        sorted_nodes = [self.criteria_names[i] for i in sorted_indices]

        n = len(sorted_nodes)
        radius = 1.0
        label_offset = 0.15

        # Evenly spaced angles from π/2 (top) to -3π/2 (full circle)
        base_angles = np.linspace(np.pi / 2, np.pi / 2 - 2 * np.pi, n, endpoint=False)

        # Alternate angles around the top for better label distribution
        angle_order = [0]
        for k in range(1, n):
            if k % 2 == 1:
                angle_order.append((k + 1) // 2)
            else:
                angle_order.append(-(k // 2))
        angles = [base_angles[(i % n)] for i in angle_order]

        # Assign positions
        circle_pos = {}
        label_pos = {}
        for node, angle in zip(sorted_nodes, angles):
            x = np.cos(angle) * radius
            y = np.sin(angle) * radius
            circle_pos[node] = (x, y)
            lx = np.cos(angle) * (radius + label_offset)
            ly = np.sin(angle) * (radius + label_offset)
            label_pos[node] = (lx, ly)

        # Draw nodes
        plt.figure(figsize=(8, 8))
        nx.draw_networkx_nodes(G, circle_pos, node_color='white', edgecolors='black', node_size=600)

        # Draw edges by style
        for style in self.edge_styles:
            edges = []
            edge_labels = {}
            for u, v, d in G.edges(data=True):
                prob = d['weight']
                lo, hi = style['range']
                if lo <= prob < hi:
                    edges.append((u, v))
                    if show_weight:
                        edge_labels[(u, v)] = f"{prob:.2f}"

            if edges:
                if show_weight:
                    nx.draw_networkx_edges(
                        G, circle_pos, edgelist=edges,
                        edge_color=style['color'],
                        width=style['width'],
                        arrows=True,
                        arrowsize=15,
                        connectionstyle='arc3,rad=0.05',
                        arrowstyle='-|>',
                    )
                    nx.draw_networkx_edge_labels(
                        G,
                        pos=circle_pos,
                        edge_labels=edge_labels,
                        font_color=style['color'],
                        font_size=9,
                        label_pos=0.5,
                        rotate=False,
                        bbox=dict(alpha=0.9, color='white', edgecolor='none', pad=0.1)
                    )
                else:
                    nx.draw_networkx_edges(
                        G, circle_pos, edgelist=edges,
                        edge_color=style['color'],
                        style=style['style'],
                        width=style['width'],
                        arrows=True,
                        arrowsize=15,
                        connectionstyle='arc3,rad=0.01',
                        arrowstyle='-|>',
                    )

        # Draw labels outside the circles, with average weights
        for node, (lx, ly) in label_pos.items():
            plt.text(lx, ly, node_labels[node], fontsize=13, ha='center', va='center',
                     bbox=dict(facecolor='white', edgecolor='none', pad=0.5))

        # Add legend
        if not show_weight:
            legend_elements = [
                Line2D([0], [0], color=style['color'], linestyle=style['style'], lw=2, label=style['label'])
                for style in self.edge_styles
            ]
            plt.legend(
                handles=legend_elements,
                title="Edge Weights",
                loc="upper center",
                bbox_to_anchor=(0.5, 1.08),
                ncol=len(legend_elements),
                frameon=False
            )
        plt.axis('off')
        plt.tight_layout()
        plt.show()

    def plot_ranking(self, **kwargs):
        '''
        Visualizes the credal ranking of criteria using a Graphviz directed graph.
        This method creates a directed graph where each node represents a criterion, labeled with its name and average weight.
        Node appearance (font size, boldness, fill color, etc.) and edge appearance (font size, boldness) can be customized via keyword arguments.
        The width of each node is dynamically estimated based on the label length for better readability.
        Edges are drawn between nodes if the probability of one criterion outranking another exceeds a threshold (default 0.5).
        For probabilities equal to 1, edges are only drawn between adjacent criteria in the sorted order of average weights.
        The resulting graph is displayed inline (e.g., in a Jupyter notebook).
        Keyword Args:
            node_fontsize (int or str): Font size for node labels (default: '14').
            edge_fontsize (int or str): Font size for edge labels (default: '14').
            node_boldfont (bool): Whether to use bold font for node labels (default: False).
            edge_boldfont (bool): Whether to use bold font for edge labels (default: False).
            graph_layout (str): Graphviz rank direction, e.g., 'TB' (top-bottom), 'LR' (left-right) (default: 'TB').
            node_fillcolor (str): Fill color for nodes (default: 'white').
        Explanation:
            - Nodes represent criteria, labeled with their names and average weights.
            - Node width is estimated based on label length for better visualization.
            - Edges are added if the probability of one criterion outranking another is greater than 0.5.
            - For probabilities exactly equal to 1, edges are only drawn between adjacent criteria in the sorted ranking.
            - Node and edge appearance can be customized via keyword arguments.
            - The graph is rendered and displayed using Graphviz.
        '''
        # Handle font size and bold fontface kwargs
        node_fontsize = kwargs.get('node_fontsize', '14')
        edge_fontsize = kwargs.get('edge_fontsize', '14')
        if isinstance(node_fontsize, int):
            node_fontsize = str(node_fontsize)
        if isinstance(edge_fontsize, int):
            edge_fontsize = str(edge_fontsize)

        node_boldfont = kwargs.get('node_boldfont', False)
        edge_boldfont = kwargs.get('edge_boldfont', False)
        node_fontname = 'Helvetica-Bold' if node_boldfont else 'Helvetica'
        edge_fontname = 'Helvetica-Bold' if edge_boldfont else 'Helvetica'

        graph_layout = kwargs.get('graph_layout', 'TB')
        node_fillcolor = kwargs.get('node_fillcolor', 'white')

        def estimate_width(label, char_width=0.13, min_width=1, max_width=3):
            width = max(min_width, min(max_width, len(label) * char_width))
            return width
        
        # You can increase the font size of node labels and edge labels in Graphviz by setting the 'fontsize' attribute.
        dot = Digraph(format='png')
        dot.attr(rankdir=graph_layout, size='8,10')

        # Add nodes with average weights as labels and flexible width
        for i, name in enumerate(self.criteria_names):
            label = f"{name}\n({self.avg_weights[i]:.2f})"
            width = estimate_width(label)
            dot.node(
                name,
                label,
                shape='ellipse',
                style='filled',
                fillcolor=node_fillcolor,
                width=str(width),
                fixedsize='false',
                fontsize=node_fontsize,
                fontname=node_fontname,
            )

        # Add edges for probabilities > min_prob (excluding self-loops)
        for i, src in enumerate(self.criteria_names):
            for j, tgt in enumerate(self.criteria_names):
                if i != j and self.probs[i, j] > 0.5:
                    prob = self.probs[i, j]
                    if prob < 1:
                        dot.edge(src, tgt, label=f"{prob:.2f}")
                    elif prob == 1:
                        sorted_indices = np.argsort(self.avg_weights)
                        idx_src = sorted_indices.tolist().index(i)
                        idx_tgt = sorted_indices.tolist().index(j)
                        if abs(idx_src - idx_tgt) == 1:
                            dot.edge(
                                src, tgt,
                                label=f"{prob:.2f}",
                                font_size=edge_fontsize,
                                fontname=edge_fontname
                            )

        display(dot)



from Models.MCDMProblem import MCDMProblem
from Models.SWING.StandardSWING import StandardSWING
import numpy as np

class GaussianSWING(StandardSWING, MCDMProblem):
    
    _basic_model = """
        data { 
            int<lower=2> CNo;
            int<lower=1> DmNo;
            array[DmNo] vector[CNo] Prf_md;
            array[DmNo] vector[CNo] Prf_sigma; 
            vector<lower=0,upper=1>[CNo] e;
            real <lower=0> gamma_param;
        } 

        parameters {             
            simplex[CNo] wStar;
            real<lower=0> kappaStar;

            array[DmNo] vector<lower=0>[CNo] Prf;
        } 

        transformed parameters {
            array[DmNo] simplex[CNo] W;

            for(i in 1:DmNo)
                W[i] = Prf[i] ./ sum(Prf[i]);       

        }

        model {
            kappaStar ~ gamma(gamma_param,gamma_param);
            wStar ~ dirichlet(0.01*e);

            W ~ dirichlet(kappaStar*wStar);

            for (i in 1:DmNo)
                Prf_md[i] ~ normal(Prf[i], Prf_sigma[i]);
            
        } 
    """

    _basic_model_clustering = """
        data { 
            int<lower=2> CNo;
            int<lower=1> DmNo;
            array[DmNo] vector[CNo] Prf_md;
            array[DmNo] vector[CNo] Prf_sigma; 
            vector<lower=0,upper=1>[CNo] e;
            real <lower=0> gamma_param;

            int<lower=2> DmC;
        } 

        parameters { 
            simplex[CNo] wStar;
            real<lower=0> kappaStar;

            array[DmNo] vector<lower=0>[CNo] Prf;

            array[DmC] simplex[CNo] wc;
            array[DmC] real<lower=0> ksi; 
            array[DmNo] simplex[DmC] theta;
        } 

        transformed parameters {
            array[DmNo] simplex[CNo] W;

            for(i in 1:DmNo)
                W[i] = Prf[i] ./ sum(Prf[i]);       

        }

        model {      
            ksi ~ gamma(gamma_param, gamma_param);
            for(i in 1:DmC)
                wc[i] ~ dirichlet(0.01*e);

            for (i in 1:DmNo){
                array[DmC] real contribution;
                for(j in 1:DmC)
                     contribution[j] = log(theta[i,j]) + dirichlet_lpdf( W[i] | ksi[j]*wc[j]);
                target += log_sum_exp(contribution);

                Prf_md[i] ~ normal(Prf[i], Prf_sigma[i]);
        }
    } 
    """
    
    _basicModelSorting = """
        data { 
            int<lower=2> CNo;
            int<lower=1> DmNo;  
            vector[CNo] AB_md[DmNo]; 
            vector[CNo] AW_md[DmNo];
            vector[CNo] AB_sigma[DmNo]; 
            vector[CNo] AW_sigma[DmNo]; 
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> AltNo;
            int<lower=2> AltC;
            matrix[AltNo, CNo] Alt;
        } 

        parameters { 
            simplex[CNo] W[DmNo];
            real<lower=0> kappa[DmNo]; 

            simplex[CNo] wStar;
            real<lower=0> kappaStar;

            vector[AltC] altMu;

            vector<lower=0>[CNo] AW[DmNo];
            vector<lower=0>[CNo] AB[DmNo];
        } 

        transformed parameters {
            simplex[CNo] AB_normalized[DmNo];
            simplex[CNo] AW_normalized[DmNo];

            for(i in 1:DmNo){
                AB_normalized[i] = (e ./ AB[i]) ./ sum(e ./ AB[i]); 
                AW_normalized[i] = AW[i] ./ sum(AW[i]);
            }

            vector[AltNo] v = Alt * wStar;
            v ./= (1-v);
            v = log(v);
            array[AltNo, AltC] real<upper=0> soft_z; // log unnormalized clusters
            
            for (n in 1:AltNo)
                for (k in 1:AltC)
                    soft_z[n, k] = -log(AltC) - 0.5 * pow(altMu[k] - v[n],2);        
        }

        model {
            kappaStar ~ gamma(.01,.01);
            wStar ~ dirichlet(0.01*e);

            W ~ dirichlet(kappaStar*wStar);
            kappa ~ gamma(.01,.01);
            
            for (i in 1:DmNo){
                AB_md[i] ~ normal(AB[i], AB_sigma[i]);
                AW_md[i] ~ normal(AW[i], AW_sigma[i]);

                AW_normalized[i] ~ dirichlet(kappa[i]*W[i]);
                AB_normalized[i] ~ dirichlet(kappa[i]*W[i]);    
            }

            for (n in 1:AltNo)
                target += log_sum_exp(soft_z[n]);

        } 
    """
    
    _correlated_model = """
        data { 
            int<lower=2> CNo;
            int<lower=1> DmNo;
            array[DmNo] vector[CNo] Prf_md;
            vector[CNo] Prf_sigma[DmNo]; 
            vector<lower=0,upper=1>[CNo] e;

            vector[CNo] mu; // mean of unnormalized weight
            cov_matrix[CNo] Sigma;
            
            real<lower=0> gamma_param; // parameter for gamma distribution
            real<lower=0> sigma_coef; // parameter for sigma
        } 

        parameters { 
            vector[CNo] W_eta[DmNo]; 
            vector[CNo] wStar_eta;
            real<lower=0> kappaStar;  
            real<lower=0> kappa[DmNo]; 

            vector<lower=0>[CNo] AB[DmNo]; 
            vector<lower=0>[CNo] AW[DmNo];
        } 

        transformed parameters {
            array[DmNo] simplex[CNo] W;

            for(i in 1:DmNo)
                W[i] = Prf[i] ./ sum(Prf[i]);       

            simplex[CNo] wStar;

            wStar = softmax(wStar_eta);
            
            for(i in 1:DmNo)
                W[i] = softmax(W_eta[i]);
        } 

        model {
            wStar_eta ~ multi_normal(mu, Sigma);
                            
            for (i in 1:DmNo){
                W_eta[i] ~ multi_normal(wStar_eta, Sigma);

                Prf_md[i] ~ normal(Prf[i], Prf_sigma[i]);
    
            }
        }
    """

    _correlatedModelClustering = """
        data { 
            int<lower=2> CNo;
            int<lower=1> DmNo;  
            vector[CNo] AB_md[DmNo]; 
            vector[CNo] AW_md[DmNo]; 
            vector[CNo] AB_sigma[DmNo]; 
            vector[CNo] AW_sigma[DmNo];  
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> DmC;

            vector[CNo] mu; // mean of unnormalized weight
            cov_matrix[CNo] Sigma;
        } 

        parameters { 
            vector[CNo] W_eta[DmNo];
            real<lower=0> kappa[DmNo];

            vector[CNo] wc_eta[DmC];
            real<lower=0> ksi[DmC]; 
            simplex[DmC] theta[DmNo];

            real<lower=0> kappaStar;

            vector<lower=0>[CNo] AB[DmNo]; 
            vector<lower=0>[CNo] AW[DmNo];  
        }

        transformed parameters {
            simplex[CNo] AB_normalized[DmNo];
            simplex[CNo] AW_normalized[DmNo];

            for(i in 1:DmNo){
                AB_normalized[i] = (e ./ AB[i]) ./ sum(e ./ AB[i]); 
                AW_normalized[i] = AW[i] ./ sum(AW[i]);
            }
            
            simplex[CNo] W[DmNo];
            simplex[CNo] wc[DmC];

            for(i in 1:DmNo)
                W[i] = softmax(W_eta[i]);
            
            for(i in 1:DmC)
                wc[i] = softmax(wc_eta[i]);  
        } 

        model {
            kappa ~ gamma(.01,.01);
            ksi ~ gamma(.001, .001);
            for (i in 1:DmC)
                wc_eta[i] ~ multi_normal(mu, Sigma);
 
            for (i in 1:DmNo) {     
                real contribution[DmC]; 
                for(j in 1:DmC)
                    contribution[j] = log(theta[i,j]) + multi_normal_lpdf( W_eta[i] | wc_eta[j], Sigma);
                target += log_sum_exp(contribution);   

                AB_md[i] ~ normal(AB[i], AB_sigma[i]);
                AW_md[i] ~ normal(AW[i], AW_sigma[i]);

                AW_normalized[i] ~ dirichlet(kappa[i]*W[i]);
                AB_normalized[i] ~ dirichlet(kappa[i]*W[i]);    
            }
        }
    """

    _correlatedModelSorting = """
        data { 
            int<lower=2> CNo;
            int<lower=1> DmNo;  
            vector[CNo] AB_md[DmNo]; 
            vector[CNo] AW_md[DmNo]; 
            vector[CNo] AB_sigma[DmNo]; 
            vector[CNo] AW_sigma[DmNo]; 
            vector<lower=0,upper=1>[CNo] e;

            int<lower=2> AltNo;
            matrix[AltNo, CNo] Alt;
            int<lower=2> AltC;
            vector<lower=0,upper=1>[AltC] eAlt;

            vector[CNo] mu; // mean of unnormalized weight
            cov_matrix[CNo] Sigma;
        } 

        parameters { 
            vector[CNo] W_eta[DmNo]; 
            vector[CNo] wStar_eta;
            real<lower=0> kappaStar;  
            real<lower=0> kappa[DmNo]; 

            vector[AltC] altMu;

            vector<lower=0>[CNo] AB[DmNo]; 
            vector<lower=0>[CNo] AW[DmNo];
        } 

        transformed parameters {
            simplex[CNo] AB_normalized[DmNo];
            simplex[CNo] AW_normalized[DmNo];

            for(i in 1:DmNo){
                AB_normalized[i] = (e ./ AB[i]) ./ sum(e ./ AB[i]); 
                AW_normalized[i] = AW[i] ./ sum(AW[i]);
            }
                   
            simplex[CNo] W[DmNo];
            simplex[CNo] wStar;

            wStar = softmax(wStar_eta);
            
            for(i in 1:DmNo) {
                W[i] = softmax(W_eta[i]);
            }

            vector[AltNo] v = Alt * wStar;
            v ./= (1-v);
            v = log(v);
            array[AltNo, AltC] real<upper=0> soft_z; // log unnormalized clusters
            for (n in 1:AltNo)
                for (k in 1:AltC)
                    soft_z[n, k] = -log(AltC) - 0.5 * pow(altMu[k] - v[n],2);
                
        } 

        model {
            wStar_eta ~ multi_normal(mu, Sigma);
            kappa ~ gamma(.01,.01);
                            
            for (i in 1:DmNo){
                W_eta[i] ~ multi_normal(wStar_eta, Sigma);

                AB_md[i] ~ normal(AB[i], AB_sigma[i]);
                AW_md[i] ~ normal(AW[i], AW_sigma[i]);

                AW_normalized[i] ~ dirichlet(kappa[i]*W[i]);
                AB_normalized[i] ~ dirichlet(kappa[i]*W[i]);    
            }

            for (n in 1:AltNo)
               target += log_sum_exp(soft_z[n]);
        }
    """

    def __init__(self, Prf_md, Prf_sigma, alternatives = None, dm_cluster_number=-1, alt_sort_number=-1, num_chain=3, num_samples=1000, opt={}):
        self.Prf_MD = np.array(Prf_md)
        self.Prf_Sigma = np.array(Prf_sigma)

        MCDMProblem.__init__(self, alternatives, dm_cluster_number, alt_sort_number, num_chain, num_samples,opt)


    @property
    def input_data(self):
        data = self._get_common_data()
        data['Prf_md'] = self.Prf_MD
        data['Prf_sigma'] = self.Prf_Sigma
        data['e'] =  np.ones(self.criteria_no)

        return data
    

    @property
    def dm_no(self):
        return self.Prf_MD.shape[0]
    
    @property
    def criteria_no(self):
        return self.Prf_MD.shape[1]

    def _check_input_data(self):
        assert self.Prf_MD.shape == self.Prf_Sigma.shape, "AB_MD and AB_Sigma must be of the same size!"

        return True





from Models.BWM.StandardBWM import *
import numpy as np
from Visualization.graph import credal_ranking, weight_distribution
#from statsmodels.stats.moment_helpers import corr2cov

a_b =  np.array([
    [ 3, 4, 6, 1, 5, 2, 9, 7],
    [ 1, 2, 8, 4, 5, 3, 9, 6],
    [ 2, 2, 3, 1, 5, 5, 9, 8],
    [ 2, 1, 8, 2, 9, 3, 8, 8],
    [ 2, 4, 9, 1, 4, 3, 5, 5],
    [ 1, 2, 9, 1, 3, 5, 5, 4],
    ])

a_w =  np.array([
    [ 7, 6, 4, 9, 5, 8, 1, 3],
    [ 9, 8, 2, 5, 4, 5, 1, 3],
    [ 8, 8, 5, 9, 5, 5, 1, 2],
    [ 8, 9, 2, 8, 1, 8, 2, 2],
    [ 8, 6, 1, 9, 6, 7, 4, 4],
    [ 9, 8, 1, 9, 7, 5, 5, 6],
    ])

# A_B = np.array(
#     [[1, 0.20, 0.11, 0.37, 0.28, 0.32],
#      [1, 0.28, 0.07, 0.40, 0.17, 0.61],
#      [1, 0.26, 0.19, 0.58, 0.33, 0.68]])

# A_W = np.array([
#     [0.11, 0.41, 1, 0.13, 0.18, 0.16],
#     [0.07, 0.34, 1, 0.40, 0.25, 0.60],
#     [0.19, 0.14, 1, 0.58, 0.35, 0.80]])


# A_B_normalized = (A_B.T / np.sum(A_B,1)).T #((1 / A_B).T / np.sum(1 / A_B, 1)).T
# A_W_normalized = ((1 / A_W).T / np.sum(1 / A_W, 1)).T # (A_W.T / np.sum(A_W,1)).T

dmNo, cNo = a_w.shape
altNo = 50
x = np.random.rand(altNo // 2, cNo)
altMat = np.concatenate([x*1,x])

criteria_name = ['C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8']
bwm = StandardBWM( AB=a_b, AW=a_w)
bwm.sampling()

credal_ranking(bwm.AggregatedWeightSamples.T, criteria_name, "nx.html")
weight_distribution(bwm.AggregatedWeightSamples.T, criteria_name, cols=1, row_based=True)

## Criteria independence 
opt = {'CriteriaDependence': True }
bwm_corr = StandardBWM( AB=a_b, AW=a_w, alternatives=altMat, opt=opt)
bwm_corr.sampling()

## Criteria Dependence with Identity Matrix
opt_cor_identity = {'CriteriaDependence': False, 'Sigma': 1*np.eye(cNo) }
bwm_cor_identity = StandardBWM( AB=a_b, AW=a_w, alternatives=altMat, opt=opt_cor_identity)
bwm_cor_identity.sampling()


## Criteria Depedence with covariance 
opt_cor_cov = {'CriteriaDependence': True }
bwm_cor_cov = StandardBWM( AB=a_b, AW=a_w, alternatives=altMat, opt=opt_cor_cov)
bwm_cor_cov.sampling()

print('Ok')



from abc import ABC, abstractmethod, abstractproperty

import stan
import numpy as np

import warnings
import sys
import os
import contextlib

## this is to make the STAN models work in Jupyter notebooks
import asyncio
import nest_asyncio


class MCDMProblem(ABC):
    _basic_model = ""
    _basic_model_clustering = ""
    _basic_model_sorting = ""
    
    _correlated_model = ""
    _correlated_model_clustering = ""
    _correlated_model_sorting = ""

    _is_correlated_model = False
    _is_clustering_required = False
    _is_sorting_required = False

    def __init__(self, alternatives, dm_cluster_number, alt_sort_number, num_chain, num_samples, options={}):
        self.alternatives = alternatives
        self.dm_cluster_number = dm_cluster_number
        self.alt_sort_number = alt_sort_number
        self.num_chains = num_chain
        self.num_samples = num_samples
        self.options = options

        if self.alternatives is not None or self.options.get('Sigma') is not None:
            self._is_correlated_model = True
        if self.options.get('CriteriaDependence') == False:
            self._is_correlated_model = False

        self._is_sorting_required = True if self.alt_sort_number > 0 else False
        self._is_clustering_required = True if self.dm_cluster_number > 0 else False

    @property
    def alt_no(self):
        return 0 if isinstance(self.alternatives, type(None)) else self.alternatives.shape[0]

    @abstractproperty
    def input_data(self):
        pass

    @abstractproperty
    def dm_no(self):
        pass
    
    @abstractproperty
    def criteria_no(self):
        pass
    
    @abstractmethod
    def _check_input_data(self):
        pass

    @property
    def model(self):
        model = 'self._'

        model = model + 'correlated_model' if self._is_correlated_model else model + 'basic_model'
        model = model + '_clustering' if self._is_clustering_required else model
        model = model + '_sorting' if self._is_sorting_required else model

        print("The used model is: ", model)
        return eval(model)

    def _get_common_data(self):
        data = {}
        data['gamma_param'] = 0.01
        data['sigma_coef'] = 0.1
        data['DmNo'] = self.dm_no
        data['CNo'] = self.criteria_no
        
        if self.dm_cluster_number > 0:
            data['DmC'] = self.dm_cluster_number

        if not isinstance(self.alternatives, type(None)):
            data['Alt'] = self.alternatives
            data['AltNo'] = self.alt_no

        if self.alt_sort_number > 0:
            data['AltC'] = self.alt_sort_number
            data['eAlt'] = np.ones(self.alt_sort_number)

        if self._is_correlated_model:
            data['mu'] = 0.01 * np.ones(self.criteria_no)
            data['Sigma'] = np.cov(self.alternatives.T)
            if not isinstance(self.options.get('Sigma'), type(None)):
                data['Sigma'] = self.options.get('Sigma')
                assert data['Sigma'].shape == (self.criteria_no, self.criteria_no)

        if self.alt_sort_number > 0 and isinstance(self.alternatives, type(None)):
            raise Exception("Alternatives should be given as input for the sorting problem!")

        return data

    def sampling(self):
        if self._check_input_data:
            nest_asyncio.apply()
            asyncio.run(asyncio.sleep(1))
            with self.suppress_stan_warnings():
                posterior = stan.build(self.model, data=self.input_data, random_seed=1)
                self.samples = posterior.sample(num_chains=self.num_chains, num_samples=self.num_samples, num_warmup=100)
                self.process_samples() 
        else:
            raise Exception("The input data is not valid")

    @contextlib.contextmanager
    def suppress_stan_warnings(self):
        import io
        f = io.StringIO()
        with contextlib.redirect_stderr(f):
            yield

    # @contextlib.contextmanager
    # def suppress_stan_warnings(self):
    #     stderr_fd = sys.stderr.fileno()
    #     with os.fdopen(os.dup(stderr_fd), 'w') as old_stderr:
    #         with open(os.devnull, 'w') as devnull:
    #             os.dup2(devnull.fileno(), stderr_fd)
    #             try:
    #                 yield
    #             finally:
    #                 os.dup2(old_stderr.fileno(), stderr_fd)



    def process_samples(self):
        self.dm_weight_samples = self.samples['W']
        self.dm_weight = np.mean(self.dm_weight_samples, axis=2)

        if self._is_clustering_required:
            self.cluster_center_samples = self.samples['wc']
            self.cluster_centers = np.mean(self.cluster_center_samples, axis=2)
            self.dm_membership_samples = self.samples['theta']
            self.dm_membership = np.mean(self.dm_membership_samples, axis=2)
            
        elif self._is_sorting_required:
            self.aggregated_weight_samples = self.samples['wStar']
            self.aggregated_weight = np.mean(self.aggregated_weight_samples, axis=1)
            
            soft_z_un = np.mean(self.samples['soft_z'], axis=2)
            soft_z = np.exp(soft_z_un)
            sum_soft_z = np.sum(soft_z, axis=1).reshape((self.alt_no, 1))
            self.alternative_membership = np.divide(soft_z, sum_soft_z)
            self.alternative_sorting = np.argmax(soft_z, axis=1)

            self.alternative_values = 1 / (1 + np.exp(-self.samples['v']))

            mu_un = np.mean(self.samples['altMu'], axis=1)
            self.sorting_centers = 1 / (1 + np.exp(-mu_un))
        
        else:
            self.aggregated_weight_samples = self.samples['wStar']
            self.aggregated_weight = np.mean(self.aggregated_weight_samples, axis=1)
    
    # def exec_async(func, *args, **kwargs):
    #     with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
    #         future = executor.submit(func, *args, **kwargs)
    #         return future.result()

# class MCDMProblemOld(ABC):
#     _basic_model = ""
#     _basic_model_clustering = ""
#     _basic_model_sorting = ""

#     _correlated_model = ""
#     _correlated_model_clustering = ""
#     _correlated_model_sorting = ""

#     _is_correlated_model = False
#     _is_clustering_required = False
#     _is_sorting_required = False

#     def __init__(self, alternatives, dm_cluster_number, alt_sort_number, num_chain, num_samples, opt={}):
#         self.alternatives = alternatives
#         self.dm_cluster_no = dm_cluster_number
#         self.alt_sort_no = alt_sort_number
#         self.num_chains = num_chain
#         self.num_samples = num_samples
#         self.options = opt

#         #self._isCorrelatedModel = False if (isinstance(self.Alternatives, type(None)) and self.Options.get('CriteriaIndependence') == (False or None)) else True
#         if self.alternatives is not None or self.options.get('Sigma') is not None:
#             self._is_correlated_model = True
#         if self.options.get('CriteriaDependence') == False:
#             self._is_correlated_model = False

#         self._is_sorting_required =  True if self.alt_sort_no > 0 else False
#         self._is_clustering_required = True if self.dm_cluster_no > 0 else False


#     @property
#     def alt_no(self):
#         return 0 if isinstance(self.alternatives, type(None)) else self.alternatives.shape[0]

#     @abstractproperty
#     def input_data(self):
#         pass

#     @abstractproperty
#     def dm_no(self):
#         pass
    
#     @abstractproperty
#     def criteria_no(self):
#         pass
    
#     @abstractmethod
#     def _check_input_data(self):
#         pass

#     @property
#     def Model(self):
#         model = 'self._'

#         model = model + 'correlated_model' if self._is_correlated_model  else model + 'basic_model'
#         model = model + 'Clustering' if self._is_clustering_required else model
#         model = model + 'Sorting' if self._is_sorting_required else model 

#         print("The used model is: ", model)
#         return eval(model)

#     def _get_common_data(self):
#         data = {}
#         data['DmNo'] = self.dm_no
#         data['CNo'] = self.criteria_no

#         if self.dm_cluster_no > 0:
#             data['DmC'] = self.dm_cluster_no

#         if not isinstance(self.alternatives, type(None)):
#             data['Alt'] = self.alternatives
#             data['AltNo'] = self.alt_no

#         if self.alt_sort_no > 0:
#             data['AltC'] = self.alt_sort_no
#             data['eAlt'] = np.ones(self.alt_sort_no)

#         if self._is_correlated_model:
#             data['mu'] = 0.01 * np.ones(self.criteria_no)
#             data['Sigma'] = np.cov(self.alternatives.T) #np.eye(self.CNo) #
#             if not isinstance(self.options.get('Sigma'), type(None)):
#                 data['Sigma'] = self.options.get('Sigma')
#                 assert data['Sigma'].shape == (self.criteria_no, self.criteria_no)


#         if self.alt_sort_no > 0 and isinstance(self.alternatives, type(None)):
#             raise Exception("Alternatives should be given as input for the sorting problem!")

#         return data

#     def sampling(self):
#         if self._check_input_data():
#             posterior = stan.build(self.Model, data=self.inputData, random_seed=1)
#             self.samples = posterior.sample(num_chains=self.numChains, num_samples=self.numSamples)

#             #posterior = self.exec_async(stan.build, self.Model, data=self.inputData, random_seed=1)
#             #self.Samples = self.exec_async(posterior.sample, num_chains=self.numChains, num_samples=self.numSamples)
#             self.process_samples() 
#         else:
#             raise Exception("The input data is not valid")

#     def process_samples(self):
#         self.DmWeightSamples = self.Samples['W']
#         self.DmWeight = np.mean(self.DmWeightSamples, axis=2)

#         if self._is_clustering_required:
#             self.ClusterCenterSamples = self.samples['wc']
#             self.ClusterCenters = np.mean(self.ClusterCenterSamples, axis=2)
#             self.DmMembershipSamples = self.samples['theta']
#             self.DmMembership = np.mean(self.DmMembershipSamples, axis=2)

#         elif self._is_sorting_required:
#             self.AggregatedWeightSamples = self.Samples['wStar']
#             self.AggregatedWeight = np.mean(self.AggregatedWeightSamples, axis=1)
            
#             soft_z_un = np.mean(self.Samples['soft_z'], axis=2)
#             soft_z = np.exp(soft_z_un)
#             sum_soft_z = np.sum(soft_z, axis=1).reshape((self.alt_no,1))
#             self.AlternativeMembership = np.divide(soft_z, sum_soft_z)
#             self.AlternativeSorting = np.argmax(soft_z, axis=1)

#             self.AlternativeValues = 1 / (1 + np.exp(-self.Samples['v']))

#             mu_un = np.mean(self.Samples['altMu'], axis=1)
#             self.SortingCenters = 1 / (1 + np.exp(-mu_un))
        
#         else:
#             self.AggregatedWeightSamples = self.Samples['wStar']
#             self.AggregatedWeight = np.mean(self.AggregatedWeightSamples, axis=1)
    
#     def exec_async(func, *args, **kwargs):
#         with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
#             future = executor.submit(func, *args, **kwargs)
#             return future.result()

        

     



from setuptools import setup, find_packages

setup(
    name="BayesMCDM",
    version="0.1.1",
    description="Bayesian Multi-Criteria Decision Making Toolkit",
    author="Majid Mohammadi",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "pystan",  
    ],
)   